@@ -49,6 +49,7 @@ import org.apache.juli.logging.Log;
 import org.apache.juli.logging.LogFactory;
 import org.apache.tomcat.util.codec.binary.Base64;
 import org.apache.tomcat.util.http.MimeHeaders;
+import org.apache.tomcat.util.log.UserDataHelper;
 import org.apache.tomcat.util.net.AbstractEndpoint.Handler.SocketState;
 import org.apache.tomcat.util.net.SSLSupport;
 import org.apache.tomcat.util.net.SocketEvent;
@@ -158,6 +159,8 @@ class Http2UpgradeHandler extends AbstractStream implements InternalHttpUpgradeH
     private volatile int lastNonFinalDataPayload;
     private volatile int lastWindowUpdate;
 
+    protected final UserDataHelper userDataHelper = new UserDataHelper(log);
+
 
     public Http2UpgradeHandler(Http2Protocol protocol, Adapter adapter, Request coyoteRequest) {
         super (STREAM_ID_ZERO);
@@ -233,7 +236,12 @@ class Http2UpgradeHandler extends AbstractStream implements InternalHttpUpgradeH
                 for (int i = 0; i < settings.length % 6; i++) {
                     int id = ByteUtil.getTwoBytes(settings, i * 6);
                     long value = ByteUtil.getFourBytes(settings, (i * 6) + 2);
-                    remoteSettings.set(Setting.valueOf(id), value);
+                    Setting key = Setting.valueOf(id);
+                    if (key == Setting.UNKNOWN) {
+                        log.warn(sm.getString("connectionSettings.unknown",
+                            connectionId, Integer.toString(id), Long.toString(value)));
+                    }
+                    remoteSettings.set(key, value);
                 }
             } catch (Http2Exception e) {
                 throw new ProtocolException(
@@ -266,6 +274,10 @@ class Http2UpgradeHandler extends AbstractStream implements InternalHttpUpgradeH
             throw new ProtocolException(sm.getString("upgradeHandler.pingFailed"), ioe);
         }
 
+        // Allow streams and connection to determine timeouts
+        socketWrapper.setReadTimeout(-1);
+        socketWrapper.setWriteTimeout(-1);
+
         if (webConnection != null) {
             processStreamOnContainerThread(stream);
         }
@@ -316,31 +328,44 @@ class Http2UpgradeHandler extends AbstractStream implements InternalHttpUpgradeH
         // Might not be necessary. init() will handle that.
         init(null);
 
-
         SocketState result = SocketState.CLOSED;
 
         try {
             switch(status) {
             case OPEN_READ:
                 synchronized (socketWrapper) {
-                    if (!socketWrapper.canWrite()) {
+                    if (socketWrapper.canWrite()) {
                         // Only send a ping if there is no other data waiting to be sent.
                         // Ping manager will ensure they aren't sent too frequently.
                         pingManager.sendPing(false);
                     }
                 }
                 try {
-                    // There is data to read so use the read timeout while
-                    // reading frames ...
-                    socketWrapper.setReadTimeout(getReadTimeout());
-                    // ... and disable the connection timeout
+                    // Disable the connection timeout while frames are processed
                     setConnectionTimeout(-1);
                     while (true) {
                         try {
-                            if (!parser.readFrame(false)) {
+                            if (!parser.readFrame()) {
                                 break;
                             }
                         } catch (StreamException se) {
+                            // Log the Stream error but not necessarily all of
+                            // them
+                            UserDataHelper.Mode logMode = userDataHelper.getNextMode();
+                            if (logMode != null) {
+                                String message = sm.getString("upgradeHandler.stream.error",
+                                        connectionId, Integer.toString(se.getStreamId()));
+                                switch (logMode) {
+                                    case INFO_THEN_DEBUG:
+                                        message += sm.getString("upgradeHandler.fallToDebug");
+                                        //$FALL-THROUGH$
+                                    case INFO:
+                                        log.info(message, se);
+                                        break;
+                                    case DEBUG:
+                                        log.debug(message, se);
+                                }
+                            }
                             // Stream errors are not fatal to the connection so
                             // continue reading frames
                             Stream stream = getStream(se.getStreamId(), false);
@@ -1202,8 +1227,8 @@ class Http2UpgradeHandler extends AbstractStream implements InternalHttpUpgradeH
         }
         try {
             socketWrapper.close();
-        } catch (IOException ioe) {
-            log.debug(sm.getString("upgradeHandler.socketCloseFailed"), ioe);
+        } catch (Exception e) {
+            log.debug(sm.getString("upgradeHandler.socketCloseFailed"), e);
         }
     }
 
@@ -1579,6 +1604,15 @@ class Http2UpgradeHandler extends AbstractStream implements InternalHttpUpgradeH
         int thisRead = 0;
 
         while (len > 0) {
+            // Blocking reads use the protocol level read timeout. Non-blocking
+            // reads do not timeout. The intention is that once a frame has
+            // started to be read, the read timeout applies until it is
+            // completely read.
+            if (nextReadBlock) {
+                socketWrapper.setReadTimeout(protocol.getReadTimeout());
+            } else {
+                socketWrapper.setReadTimeout(-1);
+            }
             thisRead = socketWrapper.read(nextReadBlock, data, pos, len);
             if (thisRead == 0) {
                 if (nextReadBlock) {
@@ -1755,15 +1789,19 @@ class Http2UpgradeHandler extends AbstractStream implements InternalHttpUpgradeH
 
         increaseOverheadCount(FrameType.PRIORITY);
 
-        AbstractNonZeroStream abstractNonZeroStream = getAbstractNonZeroStream(streamId);
-        if (abstractNonZeroStream == null) {
-            abstractNonZeroStream = createRemoteStream(streamId);
-        }
-        AbstractStream parentStream = getAbstractNonZeroStream(parentStreamId);
-        if (parentStream == null) {
-            parentStream = this;
-        }
         synchronized (priorityTreeLock) {
+            // Need to look up stream and parent stream inside the lock else it
+            // is possible for a stream to be recycled before it is
+            // reprioritised. This can result in incorrect references to the
+            // non-recycled stream being retained after reprioritisation.
+            AbstractNonZeroStream abstractNonZeroStream = getAbstractNonZeroStream(streamId);
+            if (abstractNonZeroStream == null) {
+                abstractNonZeroStream = createRemoteStream(streamId);
+            }
+            AbstractStream parentStream = getAbstractNonZeroStream(parentStreamId);
+            if (parentStream == null) {
+                parentStream = this;
+            }
             abstractNonZeroStream.rePrioritise(parentStream, exclusive, weight);
         }
     }
